// PROLL.C: PianoRoll Window

// for Win98 or later.
#define _WIN32_WINDOWS 0x0410

#include <windows.h>
#include "resource.h"
#include "runtime.h"
#include "smaf.h"
#include "proll.h"

#define POSBARHEIGHT 12

typedef struct{
	UINT RollHeight;
	UINT RollWidth;
	int VisibleTrack;
	UINT CurrentPos;
	SMAF* smaf;
	LPCTSTR OptionString;
}PROLLOPTIONSTRUCT;
#define PROLLOPTION (*(PROLLOPTIONSTRUCT*)GetWindowLong(hWnd, 0))

extern COLORREF trackcolors[16];
extern UINT DrawFunc;

static void draw(HWND hWnd, HDC hdc, int width, int height)
{
	HFONT hfold;
	int scrx, scry;

	SetTextColor(hdc, RGB(0, 0, 0));
	SetBkMode(hdc, TRANSPARENT);

	FillRectangle(hdc, 0, 0, width, height, RGB(255, 255, 255));

	{
		LOGFONT lf = {0};

		lf.lfCharSet = DEFAULT_CHARSET;
		strcopy(lf.lfFaceName, "lr SVbN");
		lf.lfHeight = -12;

		hfold = SelectObject(hdc, CreateFontIndirect(&lf));
	}

	{
		SCROLLINFO vsi, hsi;

		vsi.cbSize = hsi.cbSize = sizeof(SCROLLINFO);
		vsi.fMask = hsi.fMask = SIF_POS;
		GetScrollInfo(hWnd, SB_VERT, &vsi);
		GetScrollInfo(hWnd, SB_HORZ, &hsi);

		scrx = hsi.nPos;
		scry = vsi.nPos;
	}

	// draw pianoroll
	{
		int i;
		HPEN hpold = SelectObject(hdc, CreatePen(PS_SOLID, 1, RGB(220, 220, 220)));
		COLORREF oldcolor = SetTextColor(hdc, RGB(192, 192, 192));
		EVENT* e;

		// draw horizontal grid
		for(i = 0; i < 128; i++)
		{
			MoveToEx(hdc, 0, (127 - i) * PROLLOPTION.RollHeight - scry + POSBARHEIGHT, NULL);
			LineTo(hdc, width, (127 - i) * PROLLOPTION.RollHeight - scry + POSBARHEIGHT);

			if(!(i % 12))
			{
				FillRectangle(hdc,	0, (127 - i) * PROLLOPTION.RollHeight - scry + 1 + POSBARHEIGHT,
									width, (127 - i + 1) * PROLLOPTION.RollHeight - scry + POSBARHEIGHT, RGB(240, 240, 240));
			}
		}

		// draw vertical grid
		for(i = -(scrx * (int)PROLLOPTION.RollWidth % 100); i < width * (int)PROLLOPTION.RollWidth; i += 100)
		{
			HPEN hpold;

			if(!((i + scrx * PROLLOPTION.RollWidth) % (50 * PROLLOPTION.RollWidth)))	// darker
				hpold = SelectObject(hdc, CreatePen(PS_SOLID, 1, RGB(192, 192, 192)));

			MoveToEx(hdc, i / PROLLOPTION.RollWidth, 0 + POSBARHEIGHT, NULL);
			LineTo(hdc, i / PROLLOPTION.RollWidth, height);

			if(!((i + scrx * PROLLOPTION.RollWidth) % (50 * PROLLOPTION.RollWidth)))
				DeleteObject(SelectObject(hdc, hpold));
		}

		if(PROLLOPTION.smaf)
			for(e = PROLLOPTION.smaf->events; e; e = e->next)
			{
				DWORD gtime;

				if((e->status & 0xF0) != 0x80 && (e->status & 0xF0) != 0x90)	// Not NoteON
					continue;

				//if(scrx + width < (signed)e->time)	// Over. no draw needed any more.
				//	break;

				if(PROLLOPTION.VisibleTrack != -1 && (e->status & 0x0F) != PROLLOPTION.VisibleTrack)	// NoVisible
					continue;

				if((e->status & 0xF0) == 0x80)	// NonVel
					gtime = GetVVal(e->data + 1)/* * PROLLOPTION.smaf->gbase*/;
				else
					gtime = GetVVal(e->data + 2)/* * PROLLOPTION.smaf->gbase*/;

				FillRectangle(hdc,	e->time / PROLLOPTION.RollWidth - scrx,
									(127 - e->data[0]) * PROLLOPTION.RollHeight - scry + POSBARHEIGHT,
									(e->time + gtime) / PROLLOPTION.RollWidth - scrx,
									((127 - e->data[0]) + 1) * PROLLOPTION.RollHeight - scry + POSBARHEIGHT,
									RGB(128, 64, 128));
				FillRectangle(hdc,	e->time / PROLLOPTION.RollWidth - scrx + 1,
									(127 - e->data[0]) * PROLLOPTION.RollHeight - scry + 1 + POSBARHEIGHT,
									(e->time + gtime) / PROLLOPTION.RollWidth - scrx - 1,
									((127 - e->data[0]) + 1) * PROLLOPTION.RollHeight - scry - 1 + POSBARHEIGHT,
									trackcolors[e->status & 0x0F]);
			}

		// draw horizontal grid text(ScoreLine)
		for(i = 0; i < 128; i++)
		{
			LPCTSTR oto[] = {TEXT("C"), TEXT("C#"), TEXT("D"), TEXT("D#"), TEXT("E"), TEXT("F"), TEXT("F#"), TEXT("G"), TEXT("G#"), TEXT("A"), TEXT("A#"), TEXT("B")};
			MoveToEx(hdc, 1, (127 - i) * PROLLOPTION.RollHeight - scry + 1 + POSBARHEIGHT, NULL);
			PutText(hdc, "%-2s %d=%d", oto[i % 12], i / 12 - 2, i);
		}

		SetTextColor(hdc, oldcolor);
		DeleteObject(SelectObject(hdc, hpold));
	}

	// draw PosBar
	{
		COLORREF oldcolor = SetTextColor(hdc, GetSysColor(COLOR_BTNTEXT));
		int i;

		FillRectangle(hdc, 0, 0, width, POSBARHEIGHT, GetSysColor(COLOR_BTNFACE));

		// draw vertical grid text(TimeLine)
		for(i = -(scrx * (int)PROLLOPTION.RollWidth % 100); i < width * (int)PROLLOPTION.RollWidth; i += 100)
		{
			if((i + scrx * PROLLOPTION.RollWidth) % (50 * PROLLOPTION.RollWidth))
				continue;
			MoveToEx(hdc, i / PROLLOPTION.RollWidth + 2, 0, NULL);
			PutText(hdc, "%ums", i + scrx * PROLLOPTION.RollWidth);
		}

		SetTextColor(hdc, oldcolor);
	}

	// Start
	if(PROLLOPTION.smaf && PROLLOPTION.smaf->start)
	{
		HPEN hpold = SelectObject(hdc, CreatePen(PS_SOLID, 1, RGB(0, 128, 0)));

		MoveToEx(hdc, (PROLLOPTION.smaf->start->time - PROLLOPTION.smaf->start->duration) / PROLLOPTION.RollWidth - scrx, 0, NULL);
		LineTo(hdc, (PROLLOPTION.smaf->start->time - PROLLOPTION.smaf->start->duration) / PROLLOPTION.RollWidth - scrx, height);

		DeleteObject(SelectObject(hdc, hpold));
	}
	// End
	if(PROLLOPTION.smaf && PROLLOPTION.smaf->stop)
	{
		HPEN hpold = SelectObject(hdc, CreatePen(PS_SOLID, 1, RGB(128, 0, 128)));

		MoveToEx(hdc, PROLLOPTION.smaf->stop->time / PROLLOPTION.RollWidth - scrx, 0, NULL);
		LineTo(hdc, PROLLOPTION.smaf->stop->time / PROLLOPTION.RollWidth - scrx, height);

		DeleteObject(SelectObject(hdc, hpold));
	}

	// Current
	{
		HPEN hpold = SelectObject(hdc, CreatePen(PS_SOLID, 1, RGB(255, 0, 0)));

		MoveToEx(hdc, PROLLOPTION.CurrentPos / PROLLOPTION.RollWidth - scrx, 0, NULL);
		LineTo(hdc, PROLLOPTION.CurrentPos / PROLLOPTION.RollWidth - scrx, height);

		DeleteObject(SelectObject(hdc, hpold));
	}

	if(PROLLOPTION.VisibleTrack != -1)
	{
		SIZE s;
		TCHAR ts[32];

		SaveDC(hdc);

		SetBkMode(hdc, OPAQUE);
		SetBkColor(hdc, RGB(255, 255, 255));
		SetTextColor(hdc, RGB(255, 0, 0));

		wsprintf(ts, "Viewing Track %d", PROLLOPTION.VisibleTrack + 1);

		GetTextExtentPoint32(hdc, ts, lstrlen(ts), &s);
		TextOut(hdc, width - s.cx, height - s.cy, ts, lstrlen(ts));

		RestoreDC(hdc, -1);
	}

	if(PROLLOPTION.OptionString)
	{
		SaveDC(hdc);

		SetBkMode(hdc, OPAQUE);
		SetBkColor(hdc, RGB(255, 255, 255));
		SetTextColor(hdc, RGB(0, 0, 0));

		TextOut(hdc, 0, 0, PROLLOPTION.OptionString, lstrlen(PROLLOPTION.OptionString));

		RestoreDC(hdc, -1);
	}

	DeleteObject(SelectObject(hdc, hfold));

	return;
}

LRESULT CALLBACK PRollWndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch(msg)
	{
	case WM_CREATE:
		SetWindowLong(hWnd, 0, (LONG)halloc(sizeof(PROLLOPTIONSTRUCT)));

		PROLLOPTION.RollWidth = 25;
		PROLLOPTION.RollHeight = 3;
		PROLLOPTION.VisibleTrack = -1;
		PROLLOPTION.CurrentPos = 0;
		PROLLOPTION.smaf = NULL;
		PROLLOPTION.OptionString = NULL;

		break;
	case WM_DESTROY:
		hfree((void*)GetWindowLong(hWnd, 0));
		break;
	case WM_MOUSEWHEEL:
		if(GetAsyncKeyState(VK_CONTROL) & 0x8000)
		{
			if((short)HIWORD(wParam) > 0)
				PostMessage(hWnd, WM_HSCROLL, SB_PAGELEFT, 0);
			else
				PostMessage(hWnd, WM_HSCROLL, SB_PAGERIGHT, 0);
		}else
		{
			if((short)HIWORD(wParam) > 0)
				PostMessage(hWnd, WM_VSCROLL, SB_PAGEUP, 0);
			else
				PostMessage(hWnd, WM_VSCROLL, SB_PAGEDOWN, 0);
		}
		return 0;
		// DefWindowProcMsg]ꂽƂ]Ɠ]Ŗ[v
		//break;
	case WM_HSCROLL:
	case WM_VSCROLL:
		{
			SCROLLINFO scr;

			scr.cbSize = sizeof(scr);
			scr.fMask = SIF_ALL;
			if(msg == WM_HSCROLL)
				GetScrollInfo(hWnd, SB_HORZ, &scr);
			if(msg == WM_VSCROLL)
				GetScrollInfo(hWnd, SB_VERT, &scr);

			switch(LOWORD(wParam))
			{
			//case SB_TOP:
			case SB_LEFT:
				scr.nPos = scr.nMin;
				break;
			//case SB_BOTTOM:
			case SB_RIGHT:
				scr.nPos = scr.nMax;
				break;
			//case SB_LINEUP:
			case SB_LINELEFT:
				if(scr.nMin < scr.nPos)
					scr.nPos--;
				break;
			//case SB_LINEDOWN:
			case SB_LINERIGHT:
				if(scr.nPos < scr.nMax - 1)
					scr.nPos++;
				break;
			//case SB_PAGEUP:
			case SB_PAGELEFT:
				scr.nPos -= scr.nPage;
				break;
			//case SB_PAGEDOWN:
			case SB_PAGERIGHT:
				scr.nPos += scr.nPage;
				break;
			case SB_THUMBPOSITION:
			case SB_THUMBTRACK:
				scr.nPos = HIWORD(wParam);
				break;
			}

			scr.fMask = SIF_POS;
			if(msg == WM_HSCROLL)
				SetScrollInfo(hWnd, SB_HORZ, &scr, TRUE);
			if(msg == WM_VSCROLL)
				SetScrollInfo(hWnd, SB_VERT, &scr, TRUE);

			InvalidateRect(hWnd, NULL, FALSE);
		}

		return 0;
	case WM_PAINT:
		{
			PAINTSTRUCT ps;
			HDC hdc;
			RECT rc;

			HDC hdcbg;
			HBITMAP oldbmp;
			BITMAPINFO bi = {0};

			hdc = BeginPaint(hWnd, &ps);
			GetClientRect(hWnd, &rc);

			switch(DrawFunc)
			{
			case 0:
				break;
			case 1:
				hdcbg = hdc;
				hdc = CreateCompatibleDC(hdcbg);

				bi.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
				bi.bmiHeader.biPlanes = 1;
				bi.bmiHeader.biWidth = rc.right;
				bi.bmiHeader.biHeight = rc.bottom;
				bi.bmiHeader.biBitCount = 24;
				oldbmp = SelectObject(hdc, CreateDIBSection(hdc, &bi, DIB_RGB_COLORS, NULL, NULL, 0));

				break;
			case 2:
				hdcbg = hdc;
				hdc = CreateCompatibleDC(hdcbg);

				oldbmp = SelectObject(hdc, CreateCompatibleBitmap(hdcbg, rc.right, rc.bottom));

				break;
			}

			draw(hWnd, hdc, rc.right, rc.bottom);

			switch(DrawFunc)
			{
			case 0:
				break;
			case 1:
			case 2:
				BitBlt(hdcbg, 0, 0, rc.right, rc.bottom, hdc, 0, 0, SRCCOPY);

				DeleteObject(SelectObject(hdc, oldbmp));
				DeleteDC(hdc);

				break;
			}

			EndPaint(hWnd, &ps);
		}
		break;
	case WM_SIZE:
		{
			SCROLLINFO si;

			si.cbSize = sizeof(si);
			si.fMask = SIF_RANGE | SIF_PAGE | SIF_DISABLENOSCROLL;
			si.nMin = 0;
			si.nMax = 65535;
			si.nPage = 10;
			SetScrollInfo(hWnd, SB_HORZ, &si, TRUE);

			si.nMin = 0;
			si.nMax = 256 * 8;
			si.nPage = 10;
			SetScrollInfo(hWnd, SB_VERT, &si, TRUE);

			InvalidateRect(hWnd, NULL, FALSE);
		}
		break;
	case WM_USER:
		switch(LOWORD(wParam))
		{
		case PROLLWM_SCROLLTO:
			SetScrollPos(hWnd, SB_HORZ, (int)lParam / (int)PROLLOPTION.RollWidth + (short)HIWORD(wParam), TRUE);
			break;
		case PROLLWM_SETROLLWIDTH:
			PROLLOPTION.RollWidth = lParam;
			break;
		case PROLLWM_SETROLLHEIGHT:
			PROLLOPTION.RollHeight = lParam;
			break;
		case PROLLWM_SETVISIBLETRACK:
			PROLLOPTION.VisibleTrack = lParam;
			break;
		case PROLLWM_GETROLLWIDTH:
			return PROLLOPTION.RollWidth;
		case PROLLWM_GETROLLHEIGHT:
			return PROLLOPTION.RollHeight;
		case PROLLWM_GETVISIBLETRACK:
			return PROLLOPTION.VisibleTrack;
		case PROLLWM_GETSCROLL:
			return GetScrollPos(hWnd, SB_HORZ) * PROLLOPTION.RollWidth;
		case PROLLWM_SETSMAF:
			PROLLOPTION.smaf = (SMAF*)lParam;
			break;
		case PROLLWM_SETCURRENTPOS:
			PROLLOPTION.CurrentPos = lParam;
			InvalidateRect(hWnd, NULL, FALSE);
			break;
		case PROLLWM_SETOPTIONSTRING:
			PROLLOPTION.OptionString = (LPCTSTR)lParam;
			break;
		case PROLLWM_GETOPTIONSTRING:
			return (LRESULT)PROLLOPTION.OptionString;
		}
		break;
	case WM_LBUTTONDOWN:
	case WM_MBUTTONDOWN:
	case WM_RBUTTONDOWN:
	case WM_LBUTTONUP:
	case WM_MBUTTONUP:
	case WM_RBUTTONUP:
		SendMessage(GetParent(hWnd), msg, wParam, lParam - (POSBARHEIGHT << 16));
		return 0;
	}

	return DefWindowProc(hWnd, msg, wParam, lParam);
}

BOOL PRollInitialize(void)
{
	WNDCLASSEX wc = {0};

	wc.cbSize = sizeof(wc);
	wc.cbWndExtra = sizeof(PROLLOPTIONSTRUCT*);
	// Ƃ...RedrawBkgndœhԂȂȂ邩B
	//wc.hbrBackground = (HBRUSH)(COLOR_3DFACE + 1);
	wc.hCursor = LoadCursor(NULL, IDC_ARROW);
	wc.hInstance = GetModuleHandle(NULL);
	wc.hIcon = LoadIcon(wc.hInstance, MAKEINTRESOURCE(IDI_MAIN));
	wc.lpfnWndProc = PRollWndProc;
	wc.lpszClassName = "MMFToolPRollWC";
	wc.lpszMenuName = MAKEINTRESOURCE(IDR_MAIN);

	if(!RegisterClassEx(&wc))
		return FALSE;

	return TRUE;
}
